# -*- coding: utf-8 -*-

# Standard Library
from abc import ABCMeta, abstractmethod

# Vendor Library
from sqlalchemy.orm import load_only

# Project Library
from kinder.core.dict.models import UnitKind
from kinder.core.group.models import GroupStatusEnum
from kinder.core.declaration_status.models import DSS
from gisdo.alchemy_models import (
    AlchemyChildren, AlchemyDeclaration,
    AlchemyDeclarationUnit, AlchemyUnit, AlchemyGroup)
from gisdo.constants import ALL
from gisdo.utils import DateMixin, AgeDeltas


class BaseAlgorithm(object):

    __metaclass__ = ABCMeta

    def __init__(self, session, unit):

        self.unit = unit
        self.unit_mo_id = unit.unit_mo_id
        self.session = session

    def _kids(self, **kwargs):
        """
        Дети с датой рождения из age_range
        """

        age_range = kwargs.get('age_range', ALL)
        age_on_date = kwargs.get('on_date', AgeDeltas.calculate_date)
        down, up = AgeDeltas.get_category_deltas(age_range, age_on_date)
        kids = (
            self.session.query(
                AlchemyChildren).filter(
                    AlchemyChildren.date_of_birth > down,
                    AlchemyChildren.date_of_birth <= up)
            .options(load_only('id'))
            .subquery())
        return kids

    @abstractmethod
    def get_result_set(self):
        pass


class BaseApplicationAlgorithm(BaseAlgorithm, DateMixin):

    __metaclass__ = ABCMeta

    def _dou_units(self):
        """
        Достаём для каждого желаемого учреждения
        привязанное к нему учреждение
        """
        # Учреждение с типом ДОУ
        dou_units = (
            self.session.query(
                AlchemyUnit).filter(
                    AlchemyUnit.kind_id == UnitKind.DOU)
            .options(load_only('id'))
            .subquery())
        # JOIN с желаемыми учреждениями.
        return (
            self.session.query(
                AlchemyDeclarationUnit, 'unit_id')
            .join(dou_units)
            .options(load_only('id', 'ord', 'declaration_id'))
            .subquery())

    def _get_base_query(self, **kwargs):
        """
        Базовый запрос для группы показателей
        "Общее количество поданных заявлений..."
        """

        # Фильтрация по отчетному периоду.
        begin_report_period, end_report_period = self.get_report_period()

        _base_query = (
            self.session.query(
                AlchemyDeclaration, 'ord', 'unit_id')
            .filter(
                AlchemyDeclaration.date >= begin_report_period,
                AlchemyDeclaration.date <= end_report_period,
                AlchemyDeclaration.mo_id == self.unit_mo_id)
            .join(self._dou_units()).join(self._kids(**kwargs))
            .options(load_only('portal', 'status_id',
                               'work_type_id', 'id', 'children_id'))
            .subquery())
        return _base_query

    @abstractmethod
    def get_result_set(self, **kwargs):
        pass


class BaseEnrolledAlgorithm(BaseAlgorithm):

    __metaclass__ = ABCMeta

    def _get_fact_group(self):
        """
        Получаем фактические группы данного учреждения.
        """

        fact_group_this_unit = (
            self.session.query(
                AlchemyGroup).filter(
                    AlchemyGroup.status == GroupStatusEnum.FACT,
                    AlchemyGroup.unit_id == self.unit.id)
            .options(load_only(
                'id', 'work_type_id',
                'is_family', 'type_id'))
            .subquery())
        return fact_group_this_unit

    @abstractmethod
    def get_result_set(self):
        pass


class BaseQueueAlgorithm(BaseAlgorithm):
    """
    Базовый алгоритм для вычисления показателей по очереди
    """

    __metaclass__ = ABCMeta

    def __init__(self, cache, **kwargs):
        super(BaseQueueAlgorithm, self).__init__(session=None, **kwargs)

        self._cache = cache

    def _get_base_queue(self, age_range, on_date):

        result = []
        if self.unit.id in self._cache:
            result = self._cache[self.unit.id][on_date][age_range]

        return result

    def get_result_set(self, **kwargs):

        age_range = kwargs.get('age_range', ALL)
        on_date = kwargs.get('on_date', 'ON_NEW_YEAR')
        return self._get_base_queue(age_range, on_date)


class BaseQueueAlgorithmWithOutWantChangeDou(BaseQueueAlgorithm):
    """
    Базовый алгоритм без учета Желающих сменить ДОУ
    """

    def get_result_set(self, **kwargs):

        result = super(
            BaseQueueAlgorithmWithOutWantChangeDou,
            self).get_result_set(**kwargs)

        return filter(
            lambda x: x['status__code'] != DSS.WANT_CHANGE_DOU,
            result)


class BaseQueueWithDesiredDateInPeriodAlgorithm(BaseQueueAlgorithm,
                                                DateMixin):
    __metaclass__ = ABCMeta

    def get_result_set(self, **kwargs):

        result_set = super(
            BaseQueueWithDesiredDateInPeriodAlgorithm,
            self).get_result_set(
                **kwargs)

        down, up = self.get_desired_period()
        return filter(
            lambda x: x['desired_date']
            and (down <= x['desired_date'] <= up), result_set)


class BaseActualCallAlgorithm(
        BaseQueueAlgorithmWithOutWantChangeDou, DateMixin):
    """
    Базовый алгоритм для вычисления актуального спроса
    """

    __metaclass__ = ABCMeta

    def get_result_set(self, **kwargs):

        qs = super(BaseActualCallAlgorithm, self).get_result_set(
            **kwargs)
        return filter(
            lambda x: not x['desired_date']
            or x['desired_date'] <= self.get_desired_date(),
            qs)


class BaseDeferredQueueAlgorithm(
        BaseQueueAlgorithmWithOutWantChangeDou, DateMixin):
    """
    Базовый алгоритм для вычисления отложенного спроса
    """

    __metaclass__ = ABCMeta

    def get_result_set(self, **kwargs):

        qs = super(BaseDeferredQueueAlgorithm, self).get_result_set(
            **kwargs)
        return filter(
            lambda x: x['desired_date']
            and x['desired_date'] > self.get_desired_date(),
            qs)


class BaseHealthNeedAlgorithm(BaseQueueAlgorithm, DateMixin):

    __metaclass__ = ABCMeta

    @abstractmethod
    def _get_heath_needs(self):
        """
        Возвращает ID конкретной потребности по здоровью.
        """

    def get_result_set(self, **kwargs):

        qs = super(BaseHealthNeedAlgorithm, self).get_result_set(**kwargs)

        hn_pks = self._get_heath_needs()

        return [rec
                for rec in qs if
                (not rec["desired_date"] or
                 rec["desired_date"] <= self.get_desired_date())
                and rec["children__health_need_id"] in hn_pks
                and rec["status__code"] in [
                    DSS.REGISTERED, DSS.PRIV_CONFIRMATING]
                and rec["children__date_of_birth"] <
                self._get_current_calendar_year()]
